from __future__ import annotations

from collections import OrderedDict

import pycparser.plyparser
from angr import sim_type
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QDialog, QDialogButtonBox, QLineEdit, QMessageBox, QPushButton, QVBoxLayout

from angrmanagement.config import Conf

from .set_comment import QCommentTextBox


class CTypeEditor(QDialog):
    """
    A dialog to edit C types.

    :param parent:              The parent widget
    :param base_text:           The text to initially display in the editor
    :param multiline:           Whether the editor should be a multiline editor
    :param allow_multiple:      Whether to allow inputting more than one declaration separated by semicolons

    :ivar list[(Optional[str], SimType)] result:
                                The entered types. Will only be one entry if allow_multiple is false, and will be empty
                                if the dialog was cancelled.
    """

    def __init__(
        self,
        parent,
        arch,
        base_text: str = "",
        multiline: bool = False,
        allow_multiple: bool = False,
        predefined_types=None,
    ) -> None:
        super().__init__(parent)

        self._allow_multiple = allow_multiple
        self.arch = arch
        self._predefined_types = predefined_types

        self.text = lambda: ""
        self._ok_button: QPushButton | None
        self._init_widgets(base_text, multiline)

        self.setWindowTitle("Type editor")
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowType.WindowContextHelpButtonHint)

        self.result = []

    def _init_widgets(self, base_text, multiline) -> None:
        buttons = QDialogButtonBox(parent=self)
        buttons.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)
        buttons.accepted.connect(self._on_ok_pressed)
        buttons.rejected.connect(self._on_cancel_pressed)
        self._ok_button = buttons.button(QDialogButtonBox.StandardButton.Ok)

        if multiline:
            editor = QCommentTextBox(
                parent=self, textconfirmed_callback=self._on_ok_pressed, textchanged_callback=self._evaluate
            )
            editor.setFont(Conf.disasm_font)
            self.text = editor.toPlainText

            editor.setPlainText(base_text)
            editor.selectAll()

        else:
            editor = QLineEdit(self)
            editor.setFont(Conf.disasm_font)
            self.text = editor.text
            editor.returnPressed.connect(self._on_ok_pressed)
            editor.textChanged.connect(self._evaluate)

            editor.setText(base_text)
            editor.setFocus()
            editor.selectAll()

        layout = QVBoxLayout()
        layout.addWidget(editor)
        layout.addWidget(buttons)

        self.setLayout(layout)

    def _on_ok_pressed(self) -> None:
        if not self.result:
            return

        self.close()

    def _on_cancel_pressed(self) -> None:
        self.result = []
        self.close()

    def _evaluate(self) -> None:
        text = self.text()

        result = None
        try:
            defs, typedefs = sim_type.parse_file(text, predefined_types=self._predefined_types)
            for k in list(typedefs):
                if k.startswith("struct ") and k[7:] in typedefs:
                    typedefs.pop(k)
            result = list(typedefs.items()) + list(defs.items())
            result = [(name, ty.with_arch(self.arch)) for name, ty in result]
        except pycparser.plyparser.ParseError:
            pass

        # hack. idk why our pycparser config will accept `typedef int` as the same as `int`
        if result is None and not text.strip().startswith("typedef"):
            try:
                ty, name = sim_type.parse_type_with_name(text, predefined_types=self._predefined_types)
                result = [(name, ty.with_arch(self.arch))]
            except pycparser.plyparser.ParseError:
                pass

        if not result or (not self._allow_multiple and len(result) != 1):
            self._ok_button.setEnabled(False)
            self.result = []
        else:
            self._ok_button.setEnabled(True)
            self.result = result


def edit_field(ty, field, predefined_types=None) -> bool:
    if isinstance(ty, sim_type.SimStruct):
        fields = ty.fields
    elif isinstance(ty, sim_type.SimUnion):
        fields = ty.members
    else:
        raise TypeError(f"Can't edit a field of a {type(ty)}")

    if type(fields) is not OrderedDict:
        raise TypeError(f"Struct or union's fields are of type {type(fields)} - that's bad")
    fields_list = list(fields.items())

    if isinstance(field, int):
        if not 0 <= field < len(fields):
            raise IndexError(field)
        fieldno = field
    elif isinstance(field, str):
        try:
            fieldno = [i for i, (name, _) in fields_list if name == field][0]
        except IndexError:
            raise KeyError(field) from None
    else:
        raise TypeError(f"Field specifier is a {type(field)} - that's bad")

    name, subty = fields_list[fieldno]
    text = subty.c_repr(name=name)
    dialog = CTypeEditor(None, ty._arch, text, multiline=False, allow_multiple=False, predefined_types=predefined_types)
    dialog.exec_()
    if not dialog.result:
        return False
    name2, subty = dialog.result[0]
    if name2 is not None:
        if name != name2 and name2 in fields:
            QMessageBox.warning(None, "Duplicate field name", f"The name {name2} is already used")
        else:
            name = name2
    fields_list[fieldno] = (name, subty)

    if isinstance(ty, sim_type.SimStruct):
        ty.fields = OrderedDict(fields_list)
    elif isinstance(ty, sim_type.SimUnion):
        ty.members = OrderedDict(fields_list)
    else:
        raise TypeError(f"Can't edit a field of a {type(ty)}")

    return True



# pylint:disable=missing-class-docstring,wrong-import-order
from __future__ import annotations

import os
import sys
import unittest

import angr
from angr.analyses.decompiler.structured_codegen.c import CFunction, CFunctionCall
from common import AngrManagementTestCase, test_location
from PySide6.QtCore import Qt
from PySide6.QtTest import QTest

from angrmanagement.ui.dialogs.rename_label import RenameLabel
from angrmanagement.ui.dialogs.rename_node import RenameNode
from angrmanagement.ui.views import CodeView, DisassemblyView


class TestRenameFunctions(AngrManagementTestCase):

    def test_rename_a_function_in_disasm_and_pseudocode_views(self):
        main = self.main
        binpath = os.path.join(test_location, "x86_64", "fauxware")
        main.workspace.main_instance.project.am_obj = angr.Project(binpath, auto_load_libs=False)
        main.workspace.main_instance.project.am_event()
        main.workspace.job_manager.join_all_jobs()

        func = main.workspace.main_instance.project.kb.functions["main"]
        assert func is not None

        # decompile the function
        disasm_view = main.workspace._get_or_create_view("disassembly", DisassemblyView)
        disasm_view.display_disasm_graph()
        disasm_view.display_function(func)
        disasm_view.decompile_current_function()
        main.workspace.job_manager.join_all_jobs()

        pseudocode_view = main.workspace._get_or_create_view("pseudocode", CodeView)

        # find the node for function
        func_node = None
        assert pseudocode_view.codegen.map_pos_to_node is not None
        for _, item in pseudocode_view.codegen.map_pos_to_node.items():
            if isinstance(item.obj, CFunction):
                func_node = item.obj
                break

        assert func_node is not None, "The CFunction instance is not found."
        assert func_node.name == "main"

        # rename the function in the disassembly view
        rlabel = RenameLabel(disasm_view, func.addr, parent=None)
        rlabel._name_box.setText("")
        QTest.keyClicks(rlabel._name_box, "asdf")
        QTest.mouseClick(rlabel._ok_button, Qt.MouseButton.LeftButton)

        assert func.name == "asdf"
        assert func_node.name == "main"

        # rename the function in the pseudocode view
        rnode = RenameNode(code_view=pseudocode_view, node=func_node)
        rnode._name_box.setText("")
        QTest.keyClicks(rnode._name_box, "fdsa")
        QTest.mouseClick(rnode._ok_button, Qt.MouseButton.LeftButton)

        assert func.name == "fdsa"

    def test_rename_a_callee_in_pseudocode_view(self):
        main = self.main
        binpath = os.path.join(test_location, "x86_64", "fauxware")
        main.workspace.main_instance.project.am_obj = angr.Project(binpath, auto_load_libs=False)
        main.workspace.main_instance.project.am_event()

        main.workspace.job_manager.join_all_jobs()

        func = main.workspace.main_instance.project.kb.functions["main"]
        assert func is not None

        # decompile the function
        disasm_view = main.workspace._get_or_create_view("disassembly", DisassemblyView)
        disasm_view.display_disasm_graph()
        disasm_view.display_function(func)
        disasm_view.decompile_current_function()

        main.workspace.job_manager.join_all_jobs()

        func = main.workspace.main_instance.project.kb.functions["authenticate"]
        _ = main.workspace._get_or_create_view("disassembly", DisassemblyView)
        pseudocode_view = main.workspace._get_or_create_view("pseudocode", CodeView)

        # find the node for function
        func_node = None
        assert pseudocode_view.codegen.map_pos_to_node is not None
        for _, item in pseudocode_view.codegen.map_pos_to_node.items():
            if (
                isinstance(item.obj, CFunctionCall)
                and item.obj.callee_func
                and item.obj.callee_func.name == "authenticate"
            ):
                func_node = item.obj
                break

        assert func_node is not None, "The CFunction instance is not found."

        # rename the function in the pseudocode view
        rnode = RenameNode(code_view=pseudocode_view, node=func_node)
        rnode._name_box.setText("")
        QTest.keyClicks(rnode._name_box, "authenticate_1337")
        QTest.mouseClick(rnode._ok_button, Qt.MouseButton.LeftButton)

        assert func.name == "authenticate_1337"


if __name__ == "__main__":
    unittest.main(argv=sys.argv)



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from PySide6.QtGui import QIcon

from angrmanagement.consts import IMG_LOCATION
from angrmanagement.ui.icons import icon

from .toolbar import Toolbar, ToolbarAction

if TYPE_CHECKING:
    from angrmanagement.ui.main_window import MainWindow

try:
    import archr
except ImportError:
    archr = None


class FileToolbar(Toolbar):
    def __init__(self, main_window: MainWindow) -> None:
        super().__init__(main_window, "File")

        self.actions = [
            ToolbarAction(
                icon("file-open"),
                "Open File",
                "Open a new file for analysis",
                main_window.open_file_button,
            ),
            ToolbarAction(
                QIcon(os.path.join(IMG_LOCATION, "toolbar-docker-open.png")),
                "Open Docker Target",
                "Open a file located within a docker image for analysis",
                main_window.open_docker_button,
            ),
            ToolbarAction(
                icon("file-save"),
                "Save",
                "Save angr database",
                main_window.save_database,
            ),
        ]

        if archr is None:
            self.actions.pop(1)



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from PySide6.QtCore import QSize, Qt
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QFrame,
    QGroupBox,
    QLabel,
    QListWidget,
    QListWidgetItem,
    QSplitter,
    QVBoxLayout,
)

from angrmanagement.consts import IMG_LOCATION
from angrmanagement.data.analysis_options import (
    AnalysesConfiguration,
    AnalysisOption,
    BoolAnalysisOption,
    ChoiceAnalysisOption,
    IntAnalysisOption,
    StringAnalysisOption,
)
from angrmanagement.ui.widgets.qproperty_editor import (
    BoolPropertyItem,
    ComboPropertyItem,
    GroupPropertyItem,
    IntPropertyItem,
    PropertyModel,
    QPropertyEditor,
    TextPropertyItem,
)

if TYPE_CHECKING:
    from angrmanagement.ui.workspace import Workspace


def map_option_to_property(option: AnalysisOption):
    if isinstance(option, BoolAnalysisOption):
        return BoolPropertyItem(option.display_name, option.value, description=option.tooltip, option=option)
    elif isinstance(option, IntAnalysisOption):
        return IntPropertyItem(
            option.display_name,
            option.value,
            minimum=option.minimum_value if option.minimum_value is not None else -(2**31),
            maximum=option.maximum_value if option.maximum_value is not None else (2**31 - 1),
            description=option.tooltip,
            option=option,
        )
    elif isinstance(option, ChoiceAnalysisOption):
        return ComboPropertyItem(
            option.display_name,
            option.value,
            option.choices,
            description=option.tooltip,
            option=option,
        )
    elif isinstance(option, StringAnalysisOption):
        return TextPropertyItem(option.display_name, option.value, description=option.tooltip, option=option)
    else:
        raise ValueError("Mapper not implemented")


class AnalysisOptionsDialog(QDialog):
    """
    Dialog displaying available analyses and configuration options.
    """

    def __init__(self, analyses: AnalysesConfiguration, workspace: Workspace, parent=None) -> None:
        super().__init__(parent)
        self._workspace: Workspace = workspace
        self._analyses: AnalysesConfiguration = analyses
        self._init_widgets()
        self.setWindowTitle("Run Analysis")
        self.setMinimumSize(self.sizeHint())
        self.adjustSize()

    def sizeHint(self):  # pylint: disable=no-self-use
        return QSize(800, 600)

    #
    # Private methods
    #

    def _init_widgets(self) -> None:
        self.main_layout = QVBoxLayout()
        self.setLayout(self.main_layout)

        #
        # Available analyses
        #
        self._analysis_list = QListWidget()
        layout = QVBoxLayout()
        layout.addWidget(self._analysis_list)
        analyses_gbox = QGroupBox("Available Analyses")
        analyses_gbox.setLayout(layout)

        for analysis in self._analyses.analyses:
            item = QListWidgetItem(analysis.display_name, self._analysis_list)
            item.setFlags(item.flags() | Qt.ItemFlag.ItemIsUserCheckable)
            item.setCheckState(Qt.CheckState.Checked if analysis.enabled else Qt.CheckState.Unchecked)
            self._analysis_list.addItem(item)

        #
        # Analysis details
        #
        self._analysis_description_label = QLabel()
        self._analysis_description_label.setWordWrap(True)
        layout = QVBoxLayout()
        layout.addWidget(self._analysis_description_label)
        description_gbox = QGroupBox("Analysis Description")
        description_gbox.setLayout(layout)

        self._options_tree = QPropertyEditor()

        details_layout = QVBoxLayout()
        details_layout.addWidget(description_gbox)
        details_layout.addWidget(self._options_tree)

        splitter = QSplitter()
        left_frame = QFrame()
        left_layout = QVBoxLayout()
        left_layout.addWidget(analyses_gbox)
        left_frame.setLayout(left_layout)
        splitter.addWidget(left_frame)
        right_frame = QFrame()
        right_frame.setLayout(details_layout)
        splitter.addWidget(right_frame)
        splitter.setStretchFactor(0, 0)
        splitter.setStretchFactor(1, 1)

        left_margins = left_layout.contentsMargins()
        left_layout.setContentsMargins(0, 0, left_margins.right(), 0)
        details_margins = details_layout.contentsMargins()
        details_layout.setContentsMargins(details_margins.left(), 0, 0, 0)
        self.main_layout.addWidget(splitter)

        #
        # Dialog buttons
        #
        buttons = QDialogButtonBox(parent=self)
        buttons.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)
        ok_button = buttons.button(QDialogButtonBox.StandardButton.Ok)
        ok_button.setText("&Run Analysis")
        ok_button.setIcon(QIcon(os.path.join(IMG_LOCATION, "run-icon.svg")))
        buttons.accepted.connect(self._on_run_clicked)
        buttons.rejected.connect(self.reject)
        self.main_layout.addWidget(buttons)

        self._analysis_list.itemSelectionChanged.connect(self._update_item_details)
        self._analysis_list.itemChanged.connect(self._on_item_changed)

        if len(self._analyses) > 0:
            self._analysis_list.setCurrentRow(0)

        ok_button.setFocus()

    def _update_item_details(self) -> None:
        selected = self._analysis_list.selectedItems()
        if selected:
            analysis = self._analyses[self._analysis_list.indexFromItem(selected[0]).row()]
            self._analysis_description_label.setText(analysis.description)
            self._init_analysis_options_model(analysis)
        else:
            self._analysis_description_label.setText("Select analysis to view options.")

    def _init_analysis_options_model(self, analysis):
        root = GroupPropertyItem("Root")
        for option in analysis.options.values():
            root.addChild(map_option_to_property(option))
        model = PropertyModel(root)

        def on_value_changed(prop, value):
            prop.extra["option"].value = value

        model.valueChanged.connect(on_value_changed)
        self._options_tree.setModel(model)

    #
    # Event handlers
    #

    def _on_item_changed(self, item) -> None:
        analysis = self._analyses[self._analysis_list.indexFromItem(item).row()]
        analysis.enabled = item.checkState() == Qt.CheckState.Checked
        self._update_item_details()

    def _on_run_clicked(self) -> None:
        self.accept()



from __future__ import annotations

import binascii
from typing import TYPE_CHECKING

from PySide6.QtCore import Qt
from PySide6.QtGui import QAction, QContextMenuEvent, QCursor
from PySide6.QtWidgets import QAbstractItemView, QHeaderView, QMenu, QMessageBox, QTableWidget, QTableWidgetItem

if TYPE_CHECKING:
    from angr.knowledge_plugins.patches import Patch
    from PySide6.QtGui import QCloseEvent

    from angrmanagement.data.instance import Instance


class QPatchTableItem:
    """
    Item in the patch table describing a patch.
    """

    def __init__(self, patch, old_bytes) -> None:
        self.patch = patch
        self.old_bytes = old_bytes

    def widgets(self):
        patch = self.patch

        widgets = [
            QTableWidgetItem(f"{patch.addr:x}"),
            QTableWidgetItem(f"{len(patch)} bytes"),
            QTableWidgetItem(binascii.hexlify(self.old_bytes).decode("ascii") if self.old_bytes else "<unknown>"),
            QTableWidgetItem(binascii.hexlify(patch.new_bytes).decode("ascii")),
            QTableWidgetItem(patch.comment or ""),
        ]

        for w in widgets[:-1]:
            w.setFlags(w.flags() & ~Qt.ItemIsEditable)

        return widgets


class QPatchTable(QTableWidget):
    """
    Table of all patches.
    """

    HEADER = ["Address", "Size", "Old Bytes", "New Bytes", "Comment"]

    def __init__(self, instance: Instance, parent) -> None:
        super().__init__(parent)

        self.setColumnCount(len(self.HEADER))
        self.setHorizontalHeaderLabels(self.HEADER)
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.horizontalHeader().setSectionResizeMode(4, QHeaderView.ResizeMode.Stretch)
        self.verticalHeader().setVisible(False)

        self.items = []
        self.instance = instance
        self.instance.patches.am_subscribe(self._watch_patches)
        self._reloading: bool = False
        self.cellChanged.connect(self._on_cell_changed)
        self.reload()

    def _on_cell_changed(self, row: int, column: int) -> None:
        """
        Handle item change events, specifically to support editing comments.
        """
        if not self._reloading and column == 4:
            comment_text = self.item(row, column).text()
            self.items[row].patch.comment = comment_text
            self.instance.patches.am_event()

    def current_patch(self):
        selected_index = self.currentRow()
        if 0 <= selected_index < len(self.items):
            return self.items[selected_index]
        else:
            return None

    def reload(self) -> None:
        self._reloading = True
        self.clearContents()

        self.items = (
            [
                QPatchTableItem(item, self._get_bytes(self.instance.project, item.addr, len(item)))
                for item in self.instance.project.kb.patches.values()
            ]
            if not self.instance.project.am_none
            else []
        )
        items_count = len(self.items)
        self.setRowCount(items_count)

        for idx, item in enumerate(self.items):
            for i, it in enumerate(item.widgets()):
                self.setItem(idx, i, it)

        self._reloading = False

    def _watch_patches(self, **kwargs) -> None:  # pylint: disable=unused-argument
        if not self.instance.patches.am_none:
            self.reload()

    @staticmethod
    def _get_bytes(proj, addr: int, size: int):
        try:
            return proj.loader.memory.load(addr, size)
        except KeyError:
            return None

    def get_selected_patches(self) -> set[Patch]:
        """
        Get the set of selected patches.
        """
        return {self.items[idx.row()].patch for idx in self.selectedIndexes()}

    def revert_selected_patches(self) -> None:
        """
        Revert any selected patches.
        """
        dlg = QMessageBox()
        dlg.setWindowTitle("Revert patches")
        dlg.setText("Are you sure you want to revert selected patches?")
        dlg.setIcon(QMessageBox.Icon.Question)
        dlg.setStandardButtons(QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.Cancel)
        dlg.setDefaultButton(QMessageBox.StandardButton.Cancel)
        if dlg.exec_() != QMessageBox.StandardButton.Yes:
            return

        selected_patches = self.get_selected_patches()
        if selected_patches:
            for patch in selected_patches:
                self.instance.patches.remove_patch(patch.addr)
            self.instance.patches.am_event(removed=selected_patches)

    def contextMenuEvent(self, event: QContextMenuEvent) -> None:  # pylint: disable=unused-argument
        """
        Display view context menu.
        """
        mnu = QMenu(self)
        selected_patches = self.get_selected_patches()
        if len(selected_patches) > 0:
            act = QAction("Revert selected patches", mnu)
            act.triggered.connect(self.revert_selected_patches)
            mnu.addAction(act)
        mnu.exec_(QCursor.pos())

    def closeEvent(self, event: QCloseEvent) -> None:  # pylint:disable=unused-argument
        self.instance.patches.am_unsubscribe(self._watch_patches)



# pylint:disable=missing-class-docstring,wrong-import-order
from __future__ import annotations

import os
import unittest
from typing import TYPE_CHECKING

import angr
from angr.analyses.decompiler.structured_codegen.c import CVariable
from common import AngrManagementTestCase, test_location
from PySide6.QtCore import Qt
from PySide6.QtTest import QTest

from angrmanagement.ui.dialogs.rename_node import RenameNode
from angrmanagement.ui.views import DisassemblyView

if TYPE_CHECKING:
    from angrmanagement.ui.views import CodeView


class TestRenameVariables(AngrManagementTestCase):

    def setUp(self) -> None:
        super().setUp()

        binpath = os.path.join(test_location, "x86_64", "1after909")
        proj = angr.Project(binpath, auto_load_libs=False)
        self.main.workspace.main_instance.project.am_obj = proj
        self.main.workspace.main_instance.project.am_event()
        self.main.workspace.job_manager.join_all_jobs()

        self.func = proj.kb.functions["doit"]
        assert self.func is not None

        # decompile the function
        disasm_view = self.main.workspace._get_or_create_view("disassembly", DisassemblyView)
        disasm_view.display_disasm_graph()
        disasm_view.display_function(self.func)
        disasm_view.decompile_current_function()
        self.main.workspace.job_manager.join_all_jobs()

        self.code_view: CodeView = self.main.workspace.view_manager.first_view_in_category("pseudocode")

    def tearDown(self) -> None:
        super().tearDown()
        del self.code_view

    def test_rename_a_local_variable_in_pseudocode_view(self):
        # find a node for local variable
        local_var_node = None
        assert self.code_view.codegen.map_pos_to_node is not None
        for _, item in self.code_view.codegen.map_pos_to_node.items():
            if isinstance(item.obj, CVariable) and item.obj.unified_variable is not None:
                local_var_node = item.obj
                break

        assert local_var_node is not None

        # rename it
        rename_node = RenameNode(code_view=self.code_view, node=local_var_node, func=self.func)
        rename_node._name_box.setText("")
        QTest.keyClicks(rename_node._name_box, "var_abcd")
        QTest.mouseClick(rename_node._ok_button, Qt.MouseButton.LeftButton)

        assert local_var_node.unified_variable is not None
        assert local_var_node.unified_variable.name == "var_abcd"

    def test_rename_a_global_variable_in_pseudocode_view(self):
        # find a node for global variable
        global_var_node = None
        assert self.code_view.codegen.map_pos_to_node is not None
        for _, item in self.code_view.codegen.map_pos_to_node.items():
            if isinstance(item.obj, CVariable) and item.obj.variable.name == "stdout":
                global_var_node = item.obj
                break

        assert global_var_node is not None

        # rename it
        rename_node = RenameNode(code_view=self.code_view, node=global_var_node, func=self.func)
        rename_node._name_box.setText("")
        QTest.keyClicks(rename_node._name_box, "std_notout")
        QTest.mouseClick(rename_node._ok_button, Qt.MouseButton.LeftButton)

        assert global_var_node.variable.name == "std_notout"


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import itertools
from typing import TYPE_CHECKING

from .block_objects import FunctionHeader, Label, PhiVariable, Variables

if TYPE_CHECKING:
    from angr import Project


def locate_function(inst, addr: int):
    """
    Locate the function that contains the address.

    :param inst:
    :param int addr: The address.
    :return: The function object or None if address is not inside any function.
    :rtype: angr.knowledge_plugins.Function or None
    """

    if inst.cfg is None:
        return None

    functions = inst.kb.functions
    for _, function in functions.items():
        for block in function.blocks:
            if block.addr <= addr < block.addr + block.size:
                return function

    return None


def get_label_text(addr: int, kb, function=None):
    if addr in kb.labels:
        return kb.labels[addr] + ":"

    # default case
    if function is not None and addr == function.addr:
        s = []
        if function.name:
            s.append(f"{function.name}:")
        else:
            s.append(f"sub_{function.addr:x}:")
        if function.is_simprocedure:
            s.append("[SimProcedure]")
        if function.is_plt:
            s.append("[PLT]")
        return "\n".join(s)
    else:
        return f"loc_{addr:#x}:"


def get_block_objects(disasm, nodes, func_addr):
    """
    Get a list of objects to be displayed in a block in disassembly view. Objects may include instructions, stack
    variables, and labels.

    :param angr.analyses.Disassembly disasm:    The angr Disassembly Analysis instance.
    :param iterable nodes:                      A collection of CFG nodes.
    :param int func_addr:                       The function address of the current block.
    :return:                                    a list of Instruction objects and label names (strings).
    :rtype:                                     list
    """

    block_addrs = [node.addr for node in nodes]
    block_addr = block_addrs[0]
    insn_addrs = list(itertools.chain.from_iterable(disasm.block_to_insn_addrs[addr] for addr in block_addrs))

    lst = []

    variable_manager = disasm.kb.variables[func_addr]

    # function beginning
    if block_addr == func_addr:
        # function header
        try:
            func = disasm.kb.functions.get_by_addr(func_addr)
        except KeyError:
            func = None
        if func is not None and func.calling_convention is not None and func.prototype is not None:
            args = func.calling_convention.arg_locs(func.prototype)
            func_header = FunctionHeader(func.demangled_name, func.prototype, args)
            lst.append(func_header)

        # stack variables
        # filter out all stack variables
        variables = variable_manager.get_unified_variables(sort="stack")
        variables = sorted(variables, key=lambda v: v.offset)
        lst.append(Variables(variables))

    # phi variables
    phi_variables = variable_manager.get_phi_variables(block_addr)
    if phi_variables:
        for phi, variables in phi_variables.items():
            lst.append(PhiVariable(phi, variables))

    # instructions and labels
    for insn_addr in insn_addrs:
        if insn_addr != func_addr and insn_addr in disasm.kb.labels:
            lst.append(Label(insn_addr, get_label_text(insn_addr, disasm.kb)))
        lst.append(disasm.raw_result_map["instructions"][insn_addr])
        lst.extend(disasm.raw_result_map["ir"][insn_addr])

    # initial label, if there is any
    # FIXME: all labels should be generated during CFG recovery, and this step should not be necessary.
    if lst and not isinstance(lst[0], FunctionHeader | Label):
        # the first element should be a label
        lst.insert(0, Label(block_addrs[0], get_label_text(block_addrs[0], disasm.kb)))

    return lst


def get_out_branches(supernode):
    """
    Get a list of descriptors of branches going out from the supernode.

    :param SuperCFGNode supernode: The node to work with.
    :return: A list of out branch descriptors.
    :rtype: list
    """

    return supernode.out_branches


def address_to_text(addr: int, kb):
    """
    Properly convert an address to text for a label.

    :param int addr: The address to convert.
    :param angr.KnowledgeBase kb: The knowledgebase in use.
    :return: Text representation of the address.
    :rtype: str
    """

    if addr in kb.labels:
        return kb.labels[addr]

    return f"loc_{addr:#x}"


def get_out_branches_for_insn(out_branch_dict, ins_addr):
    if ins_addr not in out_branch_dict:
        return None

    out_branch_map = out_branch_dict[ins_addr]

    if len(out_branch_map) > 1:
        # if there are more than one targets, we return the union of non-default out branches
        keys = list(out_branch_map.keys())
        out_branch = None
        for k in keys:
            out_branch = out_branch_map[k].copy() if out_branch is None else out_branch.merge(out_branch_map[k])

        return out_branch

    else:
        return next(iter(out_branch_map.values()))


def fast_memory_load_pointer(project: Project, addr: int, size: int | None = None):
    try:
        return project.loader.memory.unpack_word(addr, size=size)
    except KeyError:
        return None


def string_at_addr(cfg, addr: int, project: Project, max_size: int = 50):
    try:
        mem_data = cfg.memory_data[addr]
    except KeyError:
        return None

    if mem_data.sort == "string":
        if mem_data.content is None:
            return None
        str_content = mem_data.content.decode("utf-8")
    elif mem_data.sort == "pointer-array":
        ptr = fast_memory_load_pointer(project, mem_data.address)
        try:
            next_level = cfg.memory_data[ptr]
        except KeyError:
            return None

        if next_level.sort != "string":
            return None

        if next_level.content is None:
            return None
        str_content = next_level.content.decode("utf-8")
    else:
        return None

    if str_content is not None:
        if len(str_content) > max_size:
            return '"' + filter_string_for_display(str_content[:max_size]) + '..."'
        else:
            return '"' + filter_string_for_display(str_content) + '"'
    else:
        return None


def should_display_string_label(cfg, insn_addr, project: Project):
    if insn_addr not in cfg.insn_addr_to_memory_data:
        return False

    memory_data = cfg.insn_addr_to_memory_data[insn_addr]
    if memory_data.sort == "string":
        return True
    elif memory_data.sort == "pointer-array" and memory_data.size == cfg.project.arch.bytes:
        # load the pointer
        ptr = fast_memory_load_pointer(project, memory_data.address)
        try:
            # see if the pointer is pointing to a string
            return cfg.memory_data[ptr].sort == "string"
        except KeyError:
            return False

    return False


def is_printable(ch):
    return 32 <= ch < 127


def filter_string_for_display(s):
    output = ""
    for ch in s.replace("\r", "\\r").replace("\n", "\\n").replace("\t", "\\t"):
        char = ord(ch)
        if not is_printable(char):
            ch = f"\\x{char:02x}"
        output += ch
    return output


def get_string_for_display(cfg, insn_addr, project: Project, max_size: int = 20) -> str | None:
    str_content = None

    try:
        memory_data = cfg.insn_addr_to_memory_data[insn_addr]
    except KeyError:
        return None

    if memory_data.sort == "string":
        try:
            str_content = memory_data.content.decode("utf-8")
        except UnicodeDecodeError:
            str_content = "<unicode-decoding-failure>"
    elif memory_data.sort == "pointer-array":
        ptr = fast_memory_load_pointer(project, memory_data.address)
        if ptr in cfg.memory_data:
            next_level = cfg.memory_data[ptr]
            if next_level.sort == "string":
                str_content = next_level.content.decode("utf-8")

    if str_content is not None:
        if len(str_content) > max_size:
            return '"' + filter_string_for_display(str_content[:max_size]) + '..."'
        else:
            return '"' + filter_string_for_display(str_content) + '"'
    else:
        return None


def get_comment_for_display(kb, insn_addr):
    if insn_addr in kb.comments:
        return kb.comments[insn_addr]
    else:
        return None



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6 import QtWidgets

if TYPE_CHECKING:
    from angr.analyses.data_dep import BaseDepNode

    from angrmanagement.ui.widgets.qdatadepgraph_block import QDataDepGraphBlock


class QDataDepGraphSearch(QtWidgets.QDialog):
    """
    Dialog that allows the user to search for a DepNode based on its value, architecture name, or instruction address.
    """

    def __init__(self, parent: QtWidgets.QWidget, data_dep_graph) -> None:
        super().__init__(parent)
        self.setWindowTitle("Data Dependency Node Search")
        self._data_dep_graph = data_dep_graph
        self._curr_value_text = ""
        self._curr_addr_text = ""
        self._curr_name_text = ""
        self._rel_nodes: list[QDataDepGraphBlock] = []
        self._curr_search_idx = -1

        self._name_line_edit = QtWidgets.QLineEdit(self)
        self._value_line_edit = QtWidgets.QLineEdit(self)
        self._address_line_edit = QtWidgets.QLineEdit(self)

        self._search_btn = QtWidgets.QPushButton("Search")
        self._close_btn = QtWidgets.QPushButton("Close")
        self._error_lbl = QtWidgets.QLabel(self)

        self._layout_manager = QtWidgets.QVBoxLayout(self)
        self._init_widgets()

        # Connect to slots
        self._search_btn.clicked.connect(self._on_search_click)
        self._close_btn.clicked.connect(self._on_close_click)

    def _safe_node_retrieval(self) -> QDataDepGraphBlock | None:
        if self._rel_nodes and 0 <= self._curr_search_idx < len(self._rel_nodes):
            return self._rel_nodes[self._curr_search_idx]
        else:
            return None

    def _on_close_click(self) -> None:
        curr_search_node = self._safe_node_retrieval()
        if curr_search_node:
            curr_search_node.selected = False
        self.accept()

    def _on_search_click(self) -> None:
        """
        Iterate through matching nodes
        """

        def _node_predicate(node: BaseDepNode):
            nonlocal val_as_int
            nonlocal addr_as_int

            matches = True

            if val_as_int:
                matches &= node.value == val_as_int
            if addr_as_int:
                matches &= node.ins_addr == addr_as_int
            if self._curr_name_text:
                matches &= self._curr_name_text in str(node)
            return matches

        self._error_lbl.hide()

        if (
            self._curr_value_text != self._value_line_edit.text()
            or self._curr_addr_text != self._address_line_edit.text()
            or self._curr_name_text != self._name_line_edit.text()
        ):
            # Change in search criteria since last click, update matching nodes
            self._curr_addr_text = self._address_line_edit.text() if self._address_line_edit.text() else ""
            self._curr_value_text = self._value_line_edit.text() if self._value_line_edit.text() else ""
            self._curr_name_text = self._name_line_edit.text() if self._name_line_edit.text() else ""
            try:
                val_as_int = int(self._curr_value_text, 16) if self._curr_value_text else None
                addr_as_int = int(self._curr_addr_text, 16) if self._curr_addr_text else None
            except ValueError:
                self._error_lbl.setText("Input must be in hexadecimal format!")
                self._error_lbl.show()

            self._rel_nodes = [n for n in self._data_dep_graph.nodes if _node_predicate(n.node)]
        if len(self._rel_nodes) == 0:
            # No matching nodes, display an error
            self._error_lbl.setText("No nodes match the given search criteria.")
            self._error_lbl.show()
            return
        curr_search_node = self._safe_node_retrieval()
        if curr_search_node:
            # Deselect previous search node
            curr_search_node.selected = False

        self._curr_search_idx = (self._curr_search_idx + 1) % len(self._rel_nodes)
        search_node: QDataDepGraphBlock = self._rel_nodes[self._curr_search_idx]
        search_node.selected = True

        self._data_dep_graph.zoom(reset=True)
        self._data_dep_graph.centerOn(search_node)
        self._data_dep_graph.zoom(restore=True)
        self._data_dep_graph.refresh()

    def _init_widgets(self) -> None:
        self._error_lbl.hide()
        self._error_lbl.setStyleSheet(self._error_lbl.styleSheet() + "color: red;")

        form_layout = QtWidgets.QFormLayout()
        form_layout.addRow("Name", self._name_line_edit)
        form_layout.addRow("Value", self._value_line_edit)
        form_layout.addRow("Instruction address", self._address_line_edit)

        btn_layout = QtWidgets.QHBoxLayout()
        btn_layout.addStretch(0)
        btn_layout.addWidget(self._search_btn)
        btn_layout.addSpacing(10)
        btn_layout.addWidget(self._close_btn)
        btn_layout.addStretch(0)

        self._layout_manager.addLayout(form_layout)
        self._layout_manager.addSpacing(20)
        self._layout_manager.addLayout(btn_layout)
        self._layout_manager.addWidget(self._error_lbl)



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

from pyqodeng.core.api import CodeEdit
from pyqodeng.core.api.panel import Panel
from pyqodeng.core.api.utils import drift_color
from pyqodeng.core.panels import LineNumberPanel, Marker, MarkerPanel
from pyqodeng.core.widgets import SplittableCodeEditTabWidget
from PySide6.QtCore import QEvent, Qt
from PySide6.QtGui import QCursor, QPainter, QPen
from PySide6.QtWidgets import QInputDialog, QLineEdit, QMenu, QPlainTextEdit, QStyle, QVBoxLayout

from angrmanagement.plugins import BasePlugin
from angrmanagement.ui.views.view import InstanceView
from angrmanagement.ui.widgets.qccode_edit import ColorSchemeIDA
from angrmanagement.ui.widgets.qccode_highlighter import QCCodeHighlighter

if TYPE_CHECKING:
    from cle import Loader
    from sortedcontainers import SortedDict

    from angrmanagement.ui.views.disassembly_view import DisassemblyView
    from angrmanagement.ui.views.symexec_view import SymexecView
    from angrmanagement.ui.workspace import Workspace


class VaildLineNumberPanel(LineNumberPanel):
    """
    The color of line numbers indicates if any address associated with this line.
    The vaild lineno are black, which means we could set a breakpoint on it and
    the invaild are grey.
    """

    def __init__(self, valid_line=None) -> None:
        super().__init__()
        self._valid_line = valid_line or set()

    def paintEvent(self, event) -> None:
        # Paints the line numbers
        self._line_color_u = drift_color(self._background_brush.color(), 250)
        self._line_color_s = drift_color(self._background_brush.color(), 280)
        Panel.paintEvent(self, event)
        if not self.isVisible():
            return
        painter = QPainter(self)
        # get style options (font, size)
        width = self.width()
        height = self.editor.fontMetrics().height()
        font = self.editor.font()
        bold_font = self.editor.font()
        bold_font.setBold(True)
        pen = QPen(self._line_color_u)
        pen_selected = QPen(self._line_color_s)
        painter.setFont(font)
        # draw every visible blocks
        for top, line, _block in self.editor.visible_blocks:
            if line + 1 in self._valid_line:
                painter.setPen(pen_selected)
                painter.setFont(bold_font)
            else:
                painter.setPen(pen)
                painter.setFont(font)
            painter.drawText(-3, top, width, height, Qt.AlignmentFlag.AlignRight, str(line + 1))


class SourceCodeViewer(CodeEdit):
    """
    CodeEdit for one source code file.
    Used by SourceCodeViewerTabWidget.
    """

    viewer: SourceViewer | None = None
    current_line: int = -1

    def __init__(self, parent) -> None:
        super().__init__(parent=parent, create_default_actions=False)

        self._valid_line = None
        self.linenumber_panel = VaildLineNumberPanel()
        self.breakpoint_panel = MarkerPanel()
        self.breakpoint_panel.add_marker_requested.connect(self.add_marker_fn)
        self.breakpoint_panel.edit_marker_requested.connect(self.edit_marker_fn)
        self.breakpoint_panel.remove_marker_requested.connect(self.remove_marker_fn)

        # self.state_panel = MarkerPanel()

        self.panels.append(self.linenumber_panel)
        self.panels.append(self.breakpoint_panel)
        # self.panels.append(self.state_panel)

    def set_valid_line(self, valid_line) -> None:
        self.linenumber_panel._valid_line = valid_line
        self._valid_line = valid_line
        self.update()

    def updateState(self, state_counter) -> None:
        self.state_panel.updateState(state_counter)

    def add_find(self) -> None:
        icon = self.style().standardIcon(QStyle.StandardPixmap.SP_FileDialogContentsView)
        desc = self.viewer.add_point(self.file.path, self.current_line, "find")
        self.breakpoint_panel.add_marker(Marker(self.current_line - 1, icon, desc))

    def add_avoid(self) -> None:
        icon = self.style().standardIcon(QStyle.StandardPixmap.SP_BrowserStop)
        desc = self.viewer.add_point(self.file.path, self.current_line, "avoid")
        self.breakpoint_panel.add_marker(Marker(self.current_line - 1, icon, desc))

    def jump_to(self, addr: int) -> None:
        self.viewer.workspace.jump_to(addr)

    def add_marker_fn(self, line) -> None:
        self.current_line = line + 1
        if self.current_line not in self._valid_line:
            return
        menu = QMenu()
        menu.addAction("Add Find", self.add_find)
        menu.addAction("Add Avoid", self.add_avoid)
        address_list = self.viewer.main.line_to_addr[(self.file.path, self.current_line)]
        if address_list:
            jump_menu = menu.addMenu("Jump to")
            for addr in address_list:
                jump_menu.addAction(f"0x{addr:x}", lambda addr=addr: self.jump_to(addr))
        menu.exec_(QCursor.pos())

    def remove_marker_fn(self, line) -> None:
        self.current_line = line + 1
        self.viewer.remove_point(self.file.path, self.current_line)
        lst = self.breakpoint_panel.marker_for_line(line)
        for m in lst:
            self.breakpoint_panel.remove_marker(m)

    def edit_marker_fn(self, line) -> None:
        self.current_line = line + 1
        menu = QMenu()
        menu.addAction("Edit Condition", self.edit_cond)
        menu.exec_(QCursor.pos())

    def edit_cond(self) -> None:
        pass

    def setHighlighter(self) -> None:
        self.modes.append(QCCodeHighlighter(self.document(), color_scheme=ColorSchemeIDA()))
        QPlainTextEdit.setReadOnly(self, True)
        self.setTextInteractionFlags(
            Qt.TextInteractionFlag.TextSelectableByKeyboard | Qt.TextInteractionFlag.TextSelectableByMouse
        )

    def event(self, event):
        if event.type() == QEvent.Type.KeyPress and event.key() in (Qt.Key.Key_Slash, Qt.Key.Key_Question):
            event.accept()
            self.comment()
            return True

        return super().event(event)

    def comment(self) -> None:
        doc = self.document()
        cursor = self.textCursor()
        cursor.clearSelection()
        block = doc.findBlock(cursor.position())
        text = block.text().rsplit("//", 1)
        if len(text) == 1:
            comment = ""
            no_comment = True
        else:
            text, comment = text
            no_comment = False
        new_comment, ok = QInputDialog.getText(None, "Comment", "Comment", QLineEdit.EchoMode.Normal, comment)
        if not ok:
            return
        cursor.movePosition(cursor.EndOfLine, cursor.MoveAnchor)
        for _ in range(len(comment)):
            cursor.deletePreviousChar()
        if no_comment:
            cursor.insertText(" // ")
        cursor.insertText(new_comment)
        cursor.block()


class SourceCodeViewerTabWidget(SplittableCodeEditTabWidget):
    """
    CodeEdit for one ELF file.
    It could contain more than one file.
    """

    editors = dict.fromkeys(SourceCodeViewer.mimetypes, SourceCodeViewer)
    fallback_editor = SourceCodeViewer
    addr_to_line: SortedDict | None = None
    line_to_addr: dict | None = None
    viewer: SourceViewer | None = None
    tabs: dict[str, SourceCodeViewer] = {}

    def __init__(self, parent=None, addr_to_line: SortedDict = None, viewer=None) -> None:
        super().__init__(parent=parent)
        self.viewer = viewer
        if addr_to_line is not None:
            self.load(addr_to_line)

    def load(self, addr_to_line: SortedDict) -> None:
        self.addr_to_line = addr_to_line  # (filename, line.state.line)
        self.line_to_addr = defaultdict(list)
        self.file_list = set()
        for addr, (filename, line) in self.addr_to_line.items():
            self.file_list.add(filename)
            self.line_to_addr[(filename, line)].append(addr)
        for fn in self.file_list:
            editor: SourceCodeViewer = self.open_document(fn)
            editor.viewer = self.viewer
            editor.setHighlighter()
            valid_line = {line for (filename, line) in self.line_to_addr if filename == fn}
            editor.set_valid_line(valid_line)
            self.tabs[fn] = editor


class SourceViewer(InstanceView):
    """
    Main class of the Source Viewer Plugin
    """

    addr_to_line: SortedDict | None = None

    main: SourceCodeViewerTabWidget | None = None

    def __init__(self, workspace: Workspace) -> None:
        super().__init__("SourceViewer", workspace, "center", workspace.main_instance)
        self.base_caption = "Source Viewer"
        self.workspace = workspace
        self.instance = workspace.main_instance
        workspace.main_instance.project.am_subscribe(self.load_from_proejct)
        self._init_widgets()

    @property
    def disasm_view(self) -> DisassemblyView:
        return self.workspace.view_manager.first_view_in_category("disassembly")

    @property
    def symexec_view(self) -> SymexecView:
        return self.workspace.view_manager.first_view_in_category("symexec")

    def _init_widgets(self) -> None:
        self.main = SourceCodeViewerTabWidget()
        self.main.viewer = self
        main_layout = QVBoxLayout()
        main_layout.addWidget(self.main)
        self.setLayout(main_layout)

    def load_from_proejct(self, **kwargs) -> None:  # pylint: disable=unused-argument
        if self.instance.project.am_none:
            return
        loader: Loader = self.instance.project.loader
        if hasattr(loader.main_object, "addr_to_line") and loader.main_object.addr_to_line is not None:
            self.main.load(loader.main_object.addr_to_line)

    def add_point(self, fn, line, typ):
        symexec_view = self.symexec_view
        if not symexec_view:
            return ""
        address_list = self.main.line_to_addr[(fn, line)]

        for addr in address_list:
            if typ == "find":
                symexec_view.find_addr_in_exec(addr)
            else:
                symexec_view.avoid_addr_in_exec(addr)
        return "\n".join([(f"0x{i:x}") for i in address_list])

    def remove_point(self, fn, line) -> None:
        symexec_view = self.symexec_view
        if not symexec_view:
            return

        address_list = self.main.line_to_addr[(fn, line)]
        for addr in address_list:
            symexec_view.remove_find_addr_in_exec(addr)
            symexec_view.remove_avoid_addr_in_exec(addr)


class SourceViewerPlugin(BasePlugin):
    """
    Plugin loader
    """

    def __init__(self, workspace: Workspace) -> None:
        super().__init__(workspace)
        self.source_viewer = SourceViewer(workspace)
        workspace.default_tabs += [self.source_viewer]
        workspace.add_view(self.source_viewer)



from __future__ import annotations

from typing import TYPE_CHECKING

import qtawesome as qta
from PySide6.QtGui import QColor, QPalette

if TYPE_CHECKING:
    from PySide6.QtGui import QIcon


NAME_TO_QTAWESOME_NAME = {
    "about": "fa5s.info-circle",
    "command-palette": "ph.squares-four-light",
    "console-view": "mdi.console-line",
    "disassembly-graph": "fa5s.sitemap",
    "disassembly-linear": "msc.list-selection",
    "disassembly-view": "msc.symbol-constant",
    "docs": "mdi6.book-open-page-variant",
    "file": "mdi.file",
    "file-open": "mdi.folder-open",
    "file-save": "mdi.floppy",
    "functions-view": "mdi.function",
    "hex-view": "mdi.hexadecimal",
    "jobs-view": "fa5s.hammer",
    "log-view": "mdi.message-bulleted",
    "patches-view": "mdi.sticker-outline",
    "plugins": "mdi.puzzle-edit",
    "preferences": "fa6s.gear",
    "pseudocode-view": "msc.json",
    "run-analysis": "mdi.arrow-right-drop-circle",
    "search": "fa5s.search",
    "strings-view": "msc.symbol-string",
    "traces-view": "mdi.go-kart-track",
    "types-view": "msc.symbol-class",
}


# XXX: QtAwesome icons won't update color on palette changes, but we can
# update the QColor objects they are initialized with and that color will be
# used in new paint events.

ICON_COLOR_ROLES = {
    QPalette.ColorGroup.Active: {
        QPalette.ColorRole.PlaceholderText: QColor(),
        QPalette.ColorRole.Text: QColor(),
    },
    QPalette.ColorGroup.Disabled: {
        QPalette.ColorRole.Text: QColor(),
    },
}


ICON_COLORS = {
    "color": ICON_COLOR_ROLES[QPalette.ColorGroup.Active][QPalette.ColorRole.Text],
    "color_disabled": ICON_COLOR_ROLES[QPalette.ColorGroup.Disabled][QPalette.ColorRole.Text],
}


def icon(key, *args, **kwargs) -> QIcon | None:
    qta_name = NAME_TO_QTAWESOME_NAME.get(key)
    if "color_role" in kwargs:
        kwargs["color"] = ICON_COLOR_ROLES[QPalette.ColorGroup.Active][kwargs["color_role"]]
    return qta.icon(qta_name, *args, **kwargs) if qta_name else None


def transfer_color(dst: QColor, src: QColor) -> None:
    dst.setRgba(src.rgba())


def update_icon_colors(palette: QPalette) -> None:
    for group, roles in ICON_COLOR_ROLES.items():
        for role, dst in roles.items():
            src = palette.color(group, role)
            transfer_color(dst, src)


qta.set_global_defaults(**ICON_COLORS)



from __future__ import annotations

from typing import TYPE_CHECKING

from angr.sim_type import ALL_TYPES, SimStruct, SimUnion, TypeRef
from PySide6.QtWidgets import QFrame, QHBoxLayout, QLabel, QMessageBox, QPushButton, QScrollArea, QVBoxLayout, QWidget

from angrmanagement.data.object_container import ObjectContainer
from angrmanagement.ui.dialogs.type_editor import CTypeEditor
from angrmanagement.ui.widgets.qtypedef import QCTypeDef

from .view import FunctionView

if TYPE_CHECKING:
    from angr.knowledge_plugins.types import TypesStore
    from angr.knowledge_plugins.variables.variable_manager import VariableManagerInternal

    from angrmanagement.data.instance import Instance
    from angrmanagement.ui.workspace import Workspace


class TypesView(FunctionView):
    """
    The view that lets you modify project.kb.types. Creates a QTypeDef for each type.
    """

    def __init__(self, workspace: Workspace, default_docking_position: str, instance: Instance) -> None:
        super().__init__("types", workspace, default_docking_position, instance)

        self.base_caption = "Types"

        self._function = ObjectContainer(None, "Current function")
        self._function.am_subscribe(self.reload)

        self._layout: QVBoxLayout = None
        self._caption_label: QLabel = None
        self._init_widgets()

        # display global types by default
        self.reload()

    #
    # Properties
    #

    @property
    def current_typestore(self) -> TypesStore:
        if self._function.am_none:
            return self.instance.kb.types
        var_manager: VariableManagerInternal = self.instance.pseudocode_variable_kb.variables[self._function.addr]
        return var_manager.types

    #
    # Other methods
    #

    def _init_widgets(self) -> None:
        outer_layout = QVBoxLayout()
        scroll_area = QScrollArea()
        scroll_contents = QWidget()
        self._layout = QVBoxLayout()

        self._caption_label = QLabel()
        outer_layout.addWidget(self._caption_label)

        status_bar = QFrame()
        status_layout = QHBoxLayout()
        status_bar.setLayout(status_layout)

        new_btn = QPushButton("Add types", self)
        new_btn.clicked.connect(self._on_new_type)
        status_layout.addWidget(new_btn)

        global_btn = QPushButton("Persistent types", self)
        global_btn.clicked.connect(self._on_persistent_types_clicked)
        status_layout.addWidget(global_btn)

        scroll_area.setWidgetResizable(True)
        scroll_contents.setLayout(self._layout)
        scroll_area.setWidget(scroll_contents)
        outer_layout.addWidget(scroll_area)
        outer_layout.addWidget(status_bar)
        self.setLayout(outer_layout)

        # add a stretch to layout so elements are top-aligned
        self._layout.addStretch()
        # background color
        # TODO: Support dark mode
        # scroll_contents.setStyleSheet("background-color: white;")

    def reload(self) -> None:
        for child in list(self._layout.parent().children()):
            if type(child) is QCTypeDef:
                self._layout.takeAt(0)
                self._layout.removeWidget(child)
                child.deleteLater()

        if self.instance.project.am_none:
            self._caption_label.setText("Types View")
            return

        # update the display
        if self.function.am_none:
            self._caption_label.setText("Persistent (global) variable types")
        else:
            txt = f"Temporary (local) variable types for function {self.function.addr:#x}"
            self._caption_label.setText(txt)

        # Load persistent types or function-specific types from types store
        types_store = self.current_typestore
        for ty in types_store.iter_own():
            widget = QCTypeDef(self._layout.parent(), ty, types_store)
            self._layout.insertWidget(self._layout.count() - 1, widget)

    def _on_new_type(self) -> None:
        dialog = CTypeEditor(
            None,
            self.instance.project.arch,
            multiline=True,
            allow_multiple=True,
            predefined_types=self.instance.kb.types,
        )
        dialog.exec_()

        types_store = self.current_typestore

        for name, ty in dialog.result:
            if name is None and type(ty) in (SimStruct, SimUnion) and ty.name != "<anon>":
                name = ty.name
            if name is None:
                name = types_store.unique_type_name()
            if name.startswith("struct "):
                name = name[7:]
            elif name.startswith("union "):
                name = name[6:]
            if name in types_store:
                if name in ALL_TYPES:
                    QMessageBox.warning(None, "Redefined builtin", f"Type {name} is a builtin and cannot be redefined")
                else:
                    types_store[name].type = ty
            else:
                new_ref = TypeRef(name, ty)
                types_store[name] = new_ref

        # reload
        self.reload()

    def _on_persistent_types_clicked(self) -> None:
        self.function = None



from __future__ import annotations

from typing import TYPE_CHECKING

from angr.analyses.deobfuscator import APIObfuscationFinder

from .job import InstanceJob

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.logic.jobmanager import JobContext


class APIDeobfuscationJob(InstanceJob):
    """
    Job for deobfuscating API usage.
    """

    def __init__(self, instance: Instance, on_finish=None) -> None:
        super().__init__("API Deobfuscation", instance, on_finish=on_finish)

    def run(self, ctx: JobContext) -> None:
        self.instance.project.analyses[APIObfuscationFinder].prep(progress_callback=ctx.set_progress)(
            variable_kb=self.instance.pseudocode_variable_kb
        )

    def __repr__(self) -> str:
        return "APIDeobfuscationJob"



# pylint:disable=no-self-use
from __future__ import annotations

from PySide6.QtCore import QMargins, QRect, QSize, Qt
from PySide6.QtGui import QPainter
from PySide6.QtWidgets import QDockWidget, QFrame, QHBoxLayout, QStyle, QStyleOptionToolBar, QWidget


class ToolBarHandle(QWidget):
    """
    Paints a tool bar handle.
    """

    def __init__(self, *vargs, **kwargs) -> None:
        super().__init__(*vargs, **kwargs)
        self.orientation = False

    def paintEvent(self, _) -> None:
        painter = QPainter(self)
        painter.setBrush(self.palette().window())
        r = QRect(0, 0, self.width(), self.height()).marginsAdded(QMargins(1, 1, 1, 1))
        painter.drawRect(r)
        opt = QStyleOptionToolBar()
        style = self.style()
        if self.orientation:
            opt.state = QStyle.StateFlag.State_Horizontal
        opt.features = QStyleOptionToolBar.ToolBarFeature.Movable
        opt.toolBarArea = Qt.ToolBarArea.NoToolBarArea
        opt.rect = r
        style.drawPrimitive(QStyle.PrimitiveElement.PE_IndicatorToolBarHandle, opt, painter, self)
        painter.end()

    def sizeHint(self):
        return QSize(15, 15)


class CustomToolBar(QFrame):
    """
    Widget to contain the tool bar handle and main widget.
    """

    def __init__(self, widget) -> None:
        super().__init__()
        self.setWidget(widget)

    def setWidget(self, widget) -> None:
        layout = QHBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.handle = ToolBarHandle(parent=self)
        layout.addWidget(self.handle)
        layout.addWidget(widget)
        self.setLayout(layout)

    def sizeHint(self):
        return QSize(25, 25)


class ToolBarDockWidget(QDockWidget):
    """
    Custom tool bar using QDockWidget for better resize handling than QToolBar for large widgets.
    """

    def __init__(self, widget, title, parent) -> None:
        super().__init__(title, parent)
        self.toolbar = CustomToolBar(widget)
        self.setWidget(self.toolbar)
        self.setTitleBarWidget(self.toolbar.handle)

    def resizeEvent(self, event) -> None:
        super().resizeEvent(event)
        horizontal = self.width() > self.height()
        if horizontal:
            self.setFeatures(self.features() | QDockWidget.DockWidgetFeature.DockWidgetVerticalTitleBar)
        else:
            self.setFeatures(self.features() & ~QDockWidget.DockWidgetFeature.DockWidgetVerticalTitleBar)
        self.toolbar.handle.orientation = horizontal



from __future__ import annotations

import itertools
import json
import random
import re
import string
from collections import defaultdict
from typing import TYPE_CHECKING

import requests
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QMessageBox
from sortedcontainers import SortedDict

from angrmanagement.config import Conf
from angrmanagement.plugins.base_plugin import BasePlugin

if TYPE_CHECKING:
    from angrmanagement.ui.workspace import Workspace


class VaRec(BasePlugin):
    """
    The plugin for supporting the VaRec plugin (private for now until it is released to the public).
    """

    def __init__(self, workspace: Workspace) -> None:
        super().__init__(workspace)

        self.transitions: set[tuple[int, int]] = set()
        self.covered_blocks = SortedDict()

        self.sink_color = Qt.GlobalColor.yellow

    MENU_BUTTONS = [
        "&Infer variable names",
    ]
    INFER_VARIABLE_NAMES = 0

    def handle_click_menu(self, idx: int) -> None:
        if idx < 0 or idx >= len(self.MENU_BUTTONS):
            return

        if self.workspace.main_instance.project.am_none:
            return

        mapping = {
            VaRec.INFER_VARIABLE_NAMES: self.infer_variable_names,
        }

        mapping.get(idx)()

    @staticmethod
    def _restore_stage(view) -> None:
        # shrug
        for v in view.codegen._variable_kb.variables[view.function.addr]._unified_variables:
            m = re.match(r"@@(\S+)@@(\S+)@@", v.name)
            if m is not None:
                var_name = m.group(1)
                v.name = var_name
        # refresh the view
        view.codegen.regenerate_text()
        view.codegen.am_event()

    @staticmethod
    def randstr(n: int = 8):
        return "".join(random.choice(string.ascii_lowercase) for _ in range(n))

    def infer_variable_names(self) -> None:
        view = self.workspace._get_or_create_pseudocode_view()
        if view.codegen.am_none:
            QMessageBox.critical(
                self.workspace._main_window,
                "Error in variable name prediction",
                "Cannot predict variable names. No pseudocode exists in the pseudocode view.",
                QMessageBox.StandardButton.Ok,
            )
            return
        if view.codegen._variable_kb is None:
            QMessageBox.critical(
                self.workspace._main_window,
                "Error in variable name prediction",
                "Cannot predict variable names. The pseudocode view does not have associated variables KB.",
                QMessageBox.StandardButton.Ok,
            )
            return

        proxies = {"http": Conf.http_proxy, "https": Conf.https_proxy} if Conf.http_proxy or Conf.https_proxy else None

        for v in view.codegen._variable_kb.variables[view.function.addr]._unified_variables:
            if not v.renamed:
                v.name = f"@@{v.name}@@{VaRec.randstr()}@@"

        view.codegen.regenerate_text()
        d = {
            "code": [
                {
                    "raw_codes": [
                        view.codegen.text,
                    ]
                }
            ]
        }
        r = requests.post(f"{Conf.varec_endpoint}", data=json.dumps(d), proxies=proxies)
        try:
            result = json.loads(r.text)
        except json.JSONDecodeError:
            self._restore_stage(view)

            QMessageBox.critical(
                self.workspace._main_window,
                "Error in variable name prediction",
                "Failed to predict names for all variables involved.",
                QMessageBox.StandardButton.Ok,
            )

            return

        varname_blacklist = {"UNK", "null", "true", "false", "return", "do", "while"}
        varname_to_predicted = defaultdict(list)

        # handle failure cases
        if "code" not in result or not result["code"]:
            self._restore_stage(view)
            QMessageBox.critical(
                self.workspace._main_window,
                "Error in variable name prediction",
                "Unexpected output returned from the backend. 'code' is not found or empty.",
                QMessageBox.StandardButton.Ok,
            )
            return
        if "predictions" not in result["code"][0] or not result["code"][0]["predictions"]:
            QMessageBox.critical(
                self.workspace._main_window,
                "Error in variable name prediction",
                "Unexpected output returned from the backend. 'predictions' is not found or empty.",
                QMessageBox.StandardButton.Ok,
            )
            self._restore_stage(view)
            return
        if len(result["code"][0]["predictions"]) == 1 and isinstance(result["code"][0]["predictions"][0], str):
            QMessageBox.critical(
                self.workspace._main_window,
                "Error in variable name prediction",
                f"Prediction failed. Error: {result['code'][0]['predictions'][0]}",
                QMessageBox.StandardButton.Ok,
            )
            self._restore_stage(view)
            return

        for idx, m in enumerate(re.finditer(r"@@(\S+)@@(\S+)@@", view.codegen.text)):
            var_name = m.group(1)
            prediction = result["code"][0]["predictions"][0][idx]
            topk = prediction["top-k"]
            # remove variable names that we don't like
            filtered_topk = [item for item in topk if item["pred_name"] not in varname_blacklist]
            if filtered_topk:
                varname_to_predicted[var_name].extend(filtered_topk)

        ctrs = defaultdict(itertools.count)

        # rename them all
        used_names = set()
        for v in view.codegen._variable_kb.variables[view.function.addr]._unified_variables:
            m = re.match(r"@@(\S+)@@\S+@@", v.name)
            if m is not None:
                var_name = m.group(1)
                predicted = varname_to_predicted[var_name]
                predicted = sorted(predicted, key=lambda x: x["confidence"], reverse=True)
                v.candidate_names = {pred["pred_name"] for pred in predicted}
                for pred in predicted:
                    if pred["pred_name"] not in used_names:
                        v.name = pred["pred_name"]
                        used_names.add(v.name)
                        break
                else:
                    if predicted:
                        v.name = predicted[0]["pred_name"] + "_" + str(next(ctrs[predicted[0]["pred_name"]]))
                    else:
                        v.name = var_name  # restore the original name
        view.codegen.am_event()



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCore import QMarginsF
from PySide6.QtGui import QBrush, QColor, QCursor, QPainterPath
from PySide6.QtWidgets import (
    QGraphicsItem,
    QGraphicsSceneMouseEvent,
    QGraphicsSimpleTextItem,
    QInputDialog,
    QLineEdit,
    QMenu,
)

from angrmanagement.config import Conf

if TYPE_CHECKING:
    from angrmanagement.ui.views.disassembly_view import DisassemblyView
    from angrmanagement.ui.views.symexec_view import SymexecView

    from .qsimulation_managers import QSimulationManagers


class QInstructionAnnotation(QGraphicsSimpleTextItem):
    """
    Abstract Instruction Annotation Class.
    It must have address prop to show at the right place.
    """

    background_color = None
    foreground_color = None
    addr = None

    @property
    def disasm_view(self) -> DisassemblyView:
        return self.parentItem().disasm_view

    @property
    def symexec_view(self) -> SymexecView:
        return self.parentItem().disasm_view.workspace.view_manager.first_view_in_category("symexec")

    def __init__(self, addr: int, text: str, *args, **kwargs) -> None:
        super().__init__(text, *args, **kwargs)
        self.addr = addr
        self.setBrush(QBrush(self.foreground_color))
        self.setFont(Conf.disasm_font)

    def paint(self, painter, *args, **kwargs) -> None:
        margin = QMarginsF(3, 0, 3, 0)
        box = self.boundingRect().marginsAdded(margin)
        path = QPainterPath()
        path.addRoundedRect(box, 3, 3)
        painter.fillPath(path, self.background_color)
        super().paint(painter, *args, **kwargs)


class QStatsAnnotation(QInstructionAnnotation):
    """
    Abstract Stats Annotation Class.
    """

    def __init__(self, addr: int, text: str, *args, **kwargs) -> None:
        super().__init__(addr, text, *args, **kwargs)
        self.setAcceptHoverEvents(True)
        self.hovered = False

    def mousePressEvent(self, event) -> None:
        pass

    def hoverEnterEvent(self, event) -> None:  # pylint: disable=unused-argument
        self.hovered = True
        if self.disasm_view:
            self.disasm_view.redraw_current_graph()

    def hoverLeaveEvent(self, event) -> None:  # pylint: disable=unused-argument
        self.hovered = False
        if self.disasm_view:
            self.disasm_view.redraw_current_graph()

    def paint(self, painter, *args, **kwargs) -> None:
        margin = QMarginsF(7, 5, 7, 5) if self.hovered else QMarginsF(3, 0, 3, 0)
        box = self.boundingRect().marginsAdded(margin)
        path = QPainterPath()
        path.addRoundedRect(box, 5, 5)
        painter.fillPath(path, self.background_color)
        super().paint(painter, *args, **kwargs)


class QActiveCount(QStatsAnnotation):
    """
    Indicating how much active states are in these address.
    We can select/move the set of states.
    Used by execution_statistics_viewer plugin.
    """

    background_color = QColor(0, 255, 0, 30)
    foreground_color = QColor(0, 60, 0)

    def __init__(self, addr: int, states) -> None:
        super().__init__(addr, str(len(states)))
        self.states = states

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:  # pylint: disable=unused-argument
        menu = QMenu()

        def _select_states() -> None:
            if self.disasm_view:
                self.disasm_view.redraw_current_graph()
            symexec_view = self.symexec_view
            if symexec_view:
                symexec_view.select_states(self.states)
                symexec_view.workspace.raise_view(symexec_view)

        def _move_states() -> None:
            disasm_view = self.disasm_view
            symexec_view = self.symexec_view
            if disasm_view is None or symexec_view is None:
                return
            to_stash, ok = QInputDialog.getText(
                disasm_view, "Move to?", "Target Stash Name:", QLineEdit.EchoMode.Normal
            )
            if ok:
                symexec_view.current_simgr.move("active", to_stash, lambda s: s in self.states)
                symexec_view._simgrs._simgr_viewer.refresh()
                disasm_view.refresh()

        menu.addAction("Select", _select_states)
        menu.addAction("Move To", _move_states)
        menu.exec_(QCursor.pos())


class QPassthroughCount(QStatsAnnotation):
    """
    Indicating how much states passthrough address.
    Used by execution_statistics_viewer plugin.
    """

    background_color = QColor(255, 0, 0, 30)
    foreground_color = QColor(60, 0, 0)

    def __init__(self, addr: int, count: int) -> None:
        super().__init__(addr, str(count))

    # def mousePressEvent(self, event):
    #     self.symexec_view.select_states_that_passed_through(self.addr)
    #     self.disasm_view.workspace.raise_view(self.symexec_view)


class QHookAnnotation(QInstructionAnnotation):
    """
    An instruction annotation for an angr project hook.
    It is added to the annotation list by fetch_qblock_annotations and
    displays an indicator next to hooked blocks.
    """

    background_color = QColor(230, 230, 230)
    foreground_color = QColor(50, 50, 50)

    def __init__(self, addr: int, *args, **kwargs) -> None:
        super().__init__(addr, "hook", *args, **kwargs)

    def contextMenuEvent(self, event) -> None:  # pylint: disable=unused-argument
        menu = QMenu()
        menu.addAction("Delete", self.delete)
        menu.exec_(QCursor.pos())

    def delete(self) -> None:
        self.disasm_view.instance.delete_hook(self.addr)
        self.disasm_view.refresh()


class QExploreAnnotation(QInstructionAnnotation):
    """
    Abstract Class for find and avoid
    """

    background_color = None
    foreground_color = QColor(230, 230, 230)
    text = None

    def __init__(self, addr: int, qsimgrs: QSimulationManagers, *args, **kwargs) -> None:
        super().__init__(addr, self.text, *args, **kwargs)
        self.qsimgrs = qsimgrs

    def contextMenuEvent(self, event) -> None:  # pylint: disable=unused-argument
        menu = QMenu()
        menu.addAction("Delete", self.delete)
        menu.exec_(QCursor.pos())

    def delete(self):
        raise NotImplementedError


class QFindAddrAnnotation(QExploreAnnotation):
    """
    An instruction annotation for explore find address.
    It is added to the annotation list by fetch_qblock_annotations
    """

    background_color = QColor(200, 230, 100)
    foreground_color = QColor(30, 80, 30)
    text = "find"

    def delete(self) -> None:
        self.qsimgrs.remove_find_address(self.addr)
        self.disasm_view.refresh()


class QBreakAnnotation(QInstructionAnnotation):
    """
    An instruction annotation for address breakpoint.
    It is added to the annotation list by fetch_qblock_annotations
    """

    background_color = QColor(200, 10, 10)
    foreground_color = QColor(220, 220, 220)
    text = "break"

    def __init__(self, bp, *args, **kwargs) -> None:
        super().__init__(bp.addr, self.text, *args, **kwargs)
        self.bp = bp

    def contextMenuEvent(self, event) -> None:  # pylint: disable=unused-argument
        menu = QMenu()
        menu.addAction("Delete", self.delete)
        menu.exec_(QCursor.pos())

    def delete(self) -> None:
        self.disasm_view.instance.breakpoint_mgr.remove_breakpoint(self.bp)


class QAvoidAddrAnnotation(QExploreAnnotation):
    """
    An instruction annotation for explore avoid address.
    It is added to the annotation list by fetch_qblock_annotations
    """

    background_color = QColor(230, 200, 100)
    foreground_color = QColor(80, 30, 30)
    text = "avoid"

    def delete(self) -> None:
        self.qsimgrs.remove_avoid_address(self.addr)
        self.disasm_view.refresh()


class QBlockAnnotations(QGraphicsItem):
    """
    Container for all instruction annotations in a QBlock
    Used in `fetch_qblock_annotations`.
    """

    PADDING = 10

    def __init__(self, addr_to_annotations: dict[int, list[QInstructionAnnotation]], *, parent, disasm_view) -> None:
        super().__init__(parent=parent)
        self.addr_to_annotations = addr_to_annotations
        self.disasm_view: DisassemblyView = disasm_view
        max_width = 0
        for _addr, annotations_ in self.addr_to_annotations.items():
            width = sum(a.boundingRect().width() + self.PADDING for a in annotations_)
            max_width = max(max_width, width)
            for annotation in annotations:
                annotation.setParentItem(self)
        self.width = max_width
        self._init_widgets()

    def get(self, addr: int):
        return self.addr_to_annotations.get(addr)

    def paint(self, painter, *args, **kwargs) -> None:
        pass

    def boundingRect(self):
        return self.childrenBoundingRect()

    def _init_widgets(self) -> None:
        # Set the x positions of all the annotations. The y positions will be set later while laying out the
        # instructions
        for _addr, annotations_ in self.addr_to_annotations.items():
            x = self.width
            for annotation in annotations_:
                annotation.setX(x - annotation.boundingRect().width())
                x -= annotation.boundingRect().width() + self.PADDING
